import * as mc from "@minecraft/server"
const { world, system } = mc;
import { Config } from "./config.js"
import { LootChest } from "./lootchest.js"
import { Notification } from "./notifications.js"

const eqTypes = ["Feet", "Legs", "Chest", "Head", "Offhand"];

export class KnockedPlayer {
    /**@type {WeakMap<player,KnockedPlayer>} */
    static instances = new WeakMap()
    /**@type {mc.Player} */
    reviver;

    player;
    revivingProgress;
    knocked_time;
    #loop_run_id;

    isSurrending;
    surrending_progress;

    /**@param {mc.Player} player */
    constructor(player, knocked_time = 0) {

        KnockedPlayer.instances.set(player, this)
        this.knocked_time = knocked_time
        console.log(knocked_time)
        this.player = player;
        this.isSurrending = false;
        this.surrending_progress = 0;
        this.revivingProgress = 0;
        this.#loop_run_id = this.loop();

        new Notification({
            title: `@${player.name} got knocked!`,
            message: `§a@${player.name}§r is asking for help!\nat ${Config.hide_coordinates ? "?" : Object.values(player.location).map(k => k.toFixed(0)).join(' ')}`,
            icon: 'textures/ui/knocked_down',
            background: 'textures/ui/orangeBorder'
        })
            .show(world.getPlayers({ excludeNames: [player.name] }));
    }

    display() {
        const backGroundColor = this.isSurrending ? "§R§E§D" : this.reviver ? "§G§R§E" : "§O§R§A"
        const text = this.isSurrending ? "Surrending..." : this.reviver ? "Reviving..." : "Hold JUMP to surrender"
        const color = this.isSurrending ? "§c|" : this.reviver ? "§a|" : "§v|"

        const progress = this.isSurrending
            ? (this.surrending_progress / Config.surrender_time) * 35
            : this.reviver
                ? (this.revivingProgress / Config.revive_time) * 35
                : ((Config.knocked_time - this.knocked_time) / Config.knocked_time) * 35

        this.player.onScreenDisplay.setActionBar(
            `§R§M§U${backGroundColor}` +
            `${text}\n` +
            `§l${Array.from({ length: 35 }, (_, i) => i < progress ? color : "§7|").join('')}§r ${((Config.knocked_time - this.knocked_time) / 20).toFixed(0)}s\n`
        )
        this.reviver?.onScreenDisplay.setActionBar(
            `§R§M§U§G§R§E§rReviving..\n§l${Array.from({ length: 35 }, (_, i) => i < progress ? "§a|" : "§7|").join('')}`
        )
        this.player.nameTag = `§l${this.player.name}\n§l${Array.from({ length: 35 }, (_, i) => i < progress ? color : "§7|").join('')}`
    }
    animation() {
        for (let i = 0; i < 9; i++)
            this.player.inputPermissions.setPermissionCategory(i, false)

        this.player.playAnimation("animation.player.sleeping", {
            stopExpression:
                "query.is_invisible || query.is_moving || query.ground_speed || query.vertical_speed"
        })
    }
    camera() {
        const loc = this.player.getHeadLocation()
        this.player.camera.setCamera("minecraft:free", {
            location: { x: loc.x, y: loc.y - 1, z: loc.z },
            rotation: { x: 10, y: 270 }
        })
        this.player.teleport(this.player.location, {
            dimension: this.player.dimension,
            rotation: { x: 10, y: 270 }
        })
    }
    save() {
        this.player.setDynamicProperty("knocked_time", this.knocked_time)
    }
    loop() {
        const job = [
            () => {
                const nearPlayers = this.player.dimension.getEntities({
                    type: 'player',
                    location: this.player.location,
                    maxDistance: 2.5
                })
                nearPlayers.forEach(player => player.onScreenDisplay.setActionBar(
                    `§R§M§U§G§R§A§rHold SNEAK to revive ${this.player.name}`
                ))
                const reviver = nearPlayers.find(k => k.isSneaking)
                this.reviver = reviver ? world.getPlayers().find(l => l.id == reviver.id) : undefined;
            },
            () => {
                if (!this.reviver && this.player.inputInfo.getButtonState("Jump") == 'Pressed') {
                    this.isSurrending = true
                    this.surrending_progress += 1
                }
                else {
                    this.isSurrending = false
                    this.surrending_progress = 0;
                }
                system.currentTick % 20 == 0 && ["mining_fatigue", "resistance", "fire_resistance", "weakness"].forEach(effect => {
                    this.player.addEffect(effect, 21, { amplifier: 255, showParticles: false })
                })
            },
            () => {
                if (!this.reviver) {
                    this.knocked_time += 1
                    this.revivingProgress += this.revivingProgress > 0 ? -1 : 0
                }
                else {
                    this.revivingProgress += 1
                }

                this.display()
            },
            () => {
                this.animation()
            },
            () => {
                this.camera()
            },
            () => {
                this.save()
            },
            () => {
                if (this.knocked_time > Config.knocked_time) {
                    this.player.onScreenDisplay.setActionBar("§R§M§U§R§E§D§cYou Died!")
                    this.kill()
                    console.log("kill")
                }
                if (this.surrending_progress >= Config.surrender_time) {
                    this.player.onScreenDisplay.setActionBar("§R§M§U§R§E§D§cYou Died!")
                    this.kill()
                    console.log("surrender")

                }
                if (this.revivingProgress >= Config.revive_time) {
                    this.revive()
                    this.reviver.onScreenDisplay.setActionBar("§R§M§U§G§R§E§aReviving completed!")
                    this.player.onScreenDisplay.setActionBar("§R§M§U§G§R§E§aYou got revived!")
                    console.log("revived")

                }
            }
        ]
        return system.runInterval(() => {
            const run_id = system.runJob((function* () {
                for (const action of job) yield action()
                return system.clearJob(run_id)
            })())
        })
    }
    revive() {
        new Notification({
            title: `Revived §e@${this.player.name}!§r`,
            message: `§e@${this.player.name}§r got revived by §e@${this.reviver?.name ?? "Totem of undying"}!`,
            icon: 'textures/ui/revived',
            background: 'textures/ui/greenBorder'
        })
            .show(world.getPlayers({ excludeNames: [this.player.name, this.reviver?.name ?? ""] }));

        this.player.addEffect("invisibility", 3, { showParticles: false })
        this.player.nameTag = this.player.name

        system.clearRun(this.#loop_run_id)
        system.runTimeout(() => {
            this.player.camera.clear()
            for (let i = 0; i < 9; i++)
                this.player.inputPermissions.setPermissionCategory(i, true)
            KnockedPlayer.instances.delete(this.player)
            this.player.setDynamicProperty('knocked_time', undefined)
            this.player.getComponent('health').setCurrentValue(6)
        }, 3)
    }
    kill() {
        this.player.nameTag = this.player.name

        new Notification({
            title: `§e@${this.player.name}§r Gave up`,
            message: `§e@${this.player.name}§r died at ${Config.hide_coordinates ? "?" : Object.values(this.player.location).map(k => k.toFixed(0)).join(' ')}`,
            icon: 'textures/ui/death',
            background: 'textures/ui/redBorder'
        })
            .showAll()
        const inv = this.player.getComponent('inventory').container
        const eq = this.player.getComponent('equippable')
        const items = [
            ...Array.from({ length: inv.size }, (_, i) => inv.getItem(i)),
            ...eqTypes.map(k => eq.getEquipment(k))
        ]
        if (Config.loot_chest) {
            const lootChest = new LootChest()
            lootChest.create(this.player, items)
        } else {
            items.forEach(item => item && this.player.dimension.spawnItem(item, this.player.location))
        }

        this.player.runCommand('clear')

        this.player.addEffect("invisibility", 3, { showParticles: false })
        this.player.camera.clear()

        system.clearRun(this.#loop_run_id)

        if (Config.is_hardcore) {
            this.player.setGameMode(mc.GameMode.Spectator)
        } else {
            this.player.kill()
        }

        system.runTimeout(() => {
            for (let i = 0; i < 9; i++)
                this.player.inputPermissions.setPermissionCategory(i, true)
            KnockedPlayer.instances.delete(this.player)
            this.player.camera.clear()

            this.player.setDynamicProperty('knocked_time', undefined)
        }, 3)
    }
}