import * as mc from "@minecraft/server"
const { world, system } = mc;
import { LootChest } from "./lootchest.js"
import { KnockedPlayer } from "./knocked.js"


// loading properties from the player when joining

system.run(() => world.getPlayers().forEach((player) => {
    player.camera.clear()
    player.addEffect("invisibility", 3, { showParticles: false })

    const knocked_time = player?.getDynamicProperty('knocked_time')
    if (knocked_time) {
        console.log("knocked")
        KnockedPlayer.instances.has(player) || new KnockedPlayer(player, knocked_time)
    }
}))

world.afterEvents.playerSpawn.subscribe(({ player, initialSpawn }) => {
    player.camera.clear()
    player.addEffect("invisibility", 3, { showParticles: false })
    if (!initialSpawn) return;

    const knocked_time = player?.getDynamicProperty('knocked_time')
    if (knocked_time) {
        console.log("knocked")
        KnockedPlayer.instances.has(player) || new KnockedPlayer(player, knocked_time)
    }
})

// totem to revive
world.beforeEvents.itemUse.subscribe(({ source: player, itemStack }) => {
    if (itemStack.typeId != "minecraft:totem_of_undying") return;
    if (!KnockedPlayer.instances.has(player)) return;

    const knockedPlayer = KnockedPlayer.instances.get(player)
    system.run(() => {
        knockedPlayer.revive()
        player.getComponent("equippable").setEquipment("Mainhand", undefined)
    })

})
// die event while not knocked => knock down

world.afterEvents.entityDie.subscribe(({ deadEntity }) => {
    const { location, dimension } = deadEntity
    const player = world.getPlayers().find(k => k.id == deadEntity.id)

    if (!player) return;
    if (KnockedPlayer.instances.has(player)) {
        // For getting the death history
        /**@type {{location:mc.Vector3,date:number,dimension:string}[]} */
        const death_locationsObj = JSON.parse(player.getDynamicProperty("death_locations") ?? "[]")
        death_locationsObj.push({ date: Date.now(), location, dimension: dimension.id })
        player.setDynamicProperty(
            "death_locations",
            JSON.stringify(death_locationsObj)
        )
        return;
    }

    player.addTag("lifesteal:ignore")
    const run_id = system.runInterval(() => {
        if (player.getComponent("health").currentValue == 0) return;
        player.teleport(location, { dimension })
        const isSameLocation = Object.entries(player.location).every(([k, v]) => v.toFixed(1) == location[k].toFixed(1))
        const isSameDimension = player.dimension.id == dimension.id

        if (isSameLocation && isSameDimension) {
            new KnockedPlayer(player)
            player.removeTag("lifesteal:ignore")
            system.clearRun(run_id)
        }
    })
})

const eqTypes = ["Feet", "Legs", "Chest", "Head", "Offhand"];

world.afterEvents.playerInteractWithEntity.subscribe((e) => {
    const { player, target: entity } = e;
    if (entity.typeId != 'rm:loot_chest') return;
    /**@type {mc.ItemStack[]} */
    const chestItems = new LootChest().fromEntity(entity)

    const inv = player.getComponent('inventory').container
    const eq = player.getComponent('equippable')
    const currentItems = [
        ...Array.from({ length: inv.size }, (_, i) => inv.getItem(i)),
        ...eqTypes.map(k => eq.getEquipment(k))
    ]
    chestItems.forEach((item, i) => {
        i < inv.size
            ? inv.setItem(i, item)
            : eq.setEquipment(eqTypes[i - inv.size], item)
    })
    currentItems.forEach(item => {
        if (item)
            inv.emptySlotsCount > 0
                ? inv.addItem(item)
                : player.dimension.spawnItem(item, player.location)
    })
})

// medikit

world.afterEvents.playerButtonInput.subscribe(({ button, player }) => {
    if (button == mc.InputButton.Sneak && player.isSneaking) {
        const eq = player.getComponent('equippable')
        let itemStack = eq.getEquipment("Mainhand")
        if (itemStack?.typeId != 'reviveme:medikit') return;

        const health = player.getComponent("health")

        let Time = 0;

        const run_id = system.runInterval(() => {
            let itemStack = eq.getEquipment("Mainhand")
            if (itemStack?.typeId != 'reviveme:medikit') return;
            const dur = itemStack.getComponent('durability')

            if (!player.isSneaking) {
                player.onScreenDisplay.setActionBar(`§R§M§U§G§R§AHold §a§lSneak§r to use the Medikit!`)
                system.clearRun(run_id)
            }
            else {
                ["mining_fatigue", "weakness"].forEach(e => {
                    player.addEffect(e, 5, { amplifier: 225, showParticles: false })
                })
                Time += 0.05;
                const time = parseInt(Time)
                if (time > 2) {
                    let { damage: progress, maxDurability: total } = dur;
                    progress = parseInt(progress / 7);
                    total = parseInt(total / 7);
                    try {
                        health.setCurrentValue(health.currentValue + 0.1);
                    } catch { }
                    try {
                        dur.damage += 1
                    } catch {
                        itemStack = null
                        system.clearRun(run_id)
                    }

                    if (Math.random() < 0.05) player.dimension.spawnParticle("minecraft:crop_growth_emitter", player.getHeadLocation())
                    player.onScreenDisplay.setActionBar(
                        `§R§M§U§G§R§EUsing Medikit...\n§l${new Array(progress).fill('§a|').concat(new Array(total - progress).fill('§7|')).join('')}`
                    )
                }
                else {
                    player.onScreenDisplay.setActionBar(
                        `§R§M§U§O§R§APreparing Medikit...\n§l${new Array(time).fill('§a------').concat(new Array(2 - time).fill('§7------')).join('')} `
                    )
                }
                if (!itemStack) {
                    player.onScreenDisplay.setActionBar(
                        `§R§M§U§G§R§ECompleted!`
                    )
                }
                eq.setEquipment("Mainhand", itemStack)
            }
        })
    }
})

// message for the medikit

world.afterEvents.playerHotbarSelectedSlotChange.subscribe(({ player, itemStack }) => {
    if (itemStack?.typeId != 'reviveme:medikit') return;
    player.onScreenDisplay.setActionBar(`§R§M§U§G§R§AHold §a§lSneak§r to use the Medikit!`)
})


// Start messages 

world.afterEvents.playerSpawn.subscribe(({ player, initialSpawn }) => {
    if (player.commandPermissionLevel == 3 && initialSpawn) {
        system.runTimeout(() => {
            player.sendMessage(
                `§a----------------------------\n` +
                `§e@${player.name}§r Thanks for playing §aRevive Me!§r\n` +
                `Bugs? §6Report them on the download page.§r\n` +
                `You like this addon? §gWrite a 5 stars review!§r\n` +
                `§aHave fun!\n` +
                `§iPack version: 1.1.4 - MC 1.21.9x\nAPI version: 2.1.0-beta\n` +
                `§a----------------------------`
            )
            player.playSound("random.levelup")
        }, 100)
    }
})

// gamerules NEEDED TO WORK
system.run(() => {
    world.gameRules.keepInventory = true;
    world.gameRules.doImmediateRespawn = true;
})
world.afterEvents.gameRuleChange.subscribe((e) => {
    if (e.rule == "keepInventory" || e.rule == "doImmediateRespawn") {
        world.gameRules.keepInventory = true;
        world.gameRules.doImmediateRespawn = true;
    }
})